"""
Copyright 2022 Objectiv B.V.
"""
import re
import string
import warnings
from typing import List

from bach.expression import Expression, join_expressions

CODES_SUPPORTED_IN_ALL_DIALECTS = {
    # These are the codes that we support for all database dialects.

    # weekday codes
    '%a',  # WEEKDAY_ABBREVIATED
    '%A',  # WEEKDAY_FULL_NAME

    # day codes
    '%d',  # DAY_OF_MONTH
    '%j',  # DAY_OF_YEAR

    # week codes
    '%V',  # ISO_8601_WEEK_NUMBER_OF_YEAR

    # month codes
    '%b',  # MONTH_ABBREVIATED
    '%B',  # MONTH_FULL_NAME
    '%m',  # MONTH_NUMBER

    # year codes
    '%y',  # YEAR_WITHOUT_CENTURY
    '%Y',  # YEAR_WITH_CENTURY

    # time unit codes
    '%H',  # HOUR24
    '%I',  # HOUR12
    '%M',  # MINUTE
    '%S',  # SECOND

    # format codes
    '%F',  # YEAR_MONTH_DAY
    '%R',  # HOUR_MINUTE
    '%T',  # HOUR_MINUTE_SECOND

    # special characters
    '%%',  # PERCENT_CHAR
}

STRINGS_SUPPORTED_IN_ALL_DIALECTS = [
    # These are the combinations of codes that we support for all database dialects, even if some of the
    # individual codes are not supported.
    # When adding more strings here: Strings must be sorted longest to shortest, and all strings must
    # start with a percentage sign.

    '%S.%f',  # <seconds>.<microsecnds>
]


_SUPPORTED_C_STANDARD_CODES = {
    # week codes
    '%a',  # WEEKDAY_ABBREVIATED
    '%A',  # WEEKDAY_FULL_NAME
    '%w',  # WEEKDAY_NUMBER
    '%U',  # WEEK_NUMBER_OF_YEAR_SUNDAY_FIRST
    '%W',  # WEEK_NUMBER_OF_YEAR_MONDAY_FIRST

    # day codes
    '%d',  # DAY_OF_MONTH
    '%e',  # DAY_OF_MONTH_PRECEDED_BY_A_SPACE
    '%j',  # DAY_OF_YEAR

    # month codes
    '%b',  # MONTH_ABBREVIATED
    '%h',  # MONTH_ABBREVIATED_2
    '%B',  # MONTH_FULL_NAME
    '%m',  # MONTH_NUMBER

    # year codes
    '%y',  # YEAR_WITHOUT_CENTURY
    '%Y',  # YEAR_WITH_CENTURY
    '%C',  # CENTURY
    '%Q',  # QUARTER
    '%D',  # MONTH_DAY_YEAR

    # iso 8601 codes
    '%G',  # ISO_8601_YEAR_WITH_CENTURY
    '%g',  # ISO_8601_YEAR_WITHOUT_CENTURY
    '%V',  # ISO_8601_WEEK_NUMBER_OF_YEAR
    '%u',  # ISO_8601_WEEKDAY_NUMBER

    # time unit codes
    '%H',  # HOUR24
    '%I',  # HOUR12
    '%k',  # HOUR24_PRECEDED_BY_A_SPACE
    '%l',  # HOUR12_PRECEDED_BY_A_SPACE
    '%M',  # MINUTE
    '%S',  # SECOND
    '%s',  # EPOCH
    '%f',  # MICROSECOND

    '%z',  # UTC_OFFSET
    '%Z',  # TIME_ZONE_NAME

    # format codes
    '%F',  # YEAR_MONTH_DAY
    '%R',  # HOUR_MINUTE
    '%T',  # HOUR_MINUTE_SECOND

    # special characters
    '%n',  # NEW_LINE
    '%t',  # TAB
    '%%',  # PERCENT_CHAR
}

# https://www.postgresql.org/docs/current/functions-formatting.html#FUNCTIONS-FORMATTING-DATETIME-TABLE
_C_STANDARD_CODES_X_POSTGRES_DATE_CODES = {
    "%a": "Dy",
    "%A": "FMDay",
    "%w": "D",  # Sunday is 1 and Saturday is 7
    "%d": "DD",
    "%j": "DDD",
    "%b": "Mon",
    "%h": "Mon",
    "%B": "FMMonth",
    "%m": "MM",
    "%y": "YY",
    "%Y": "YYYY",
    "%C": "CC",
    "%Q": "Q",
    "%D": "MM/DD/YY",
    "%F": "YYYY-MM-DD",
    "%G": "IYYY",
    "%g": "IY",
    "%V": "IW",
    "%u": "ID",
    "%H": "HH24",
    "%I": "HH12",
    "%M": "MI",
    "%S": "SS",
    "%f": "US",
    "%z": "OF",
    "%Z": "TZ",
    "%R": "HH24:MI",
    "%T": "HH24:MI:SS",
    "%%": "%",
}


def parse_c_standard_code_to_postgres_code(date_format: str) -> str:
    """
    Parses a date format string from standard codes to Postgres date codes.
    If c-standard code has no equivalent, it will remain in resultant string.

    Steps to follow:
        date_format = '%Y%m-%Y%m-%Y%m%d-%d'

        Step 1: Get all unique groups of continuous c-codes,
            groups are processed based on length in order to avoid
            replacing occurrences of other groups (e.g %Y%m occurs in %Y%m%d, but both are different groups):
                ['%Y%m%d', '%Y%m', '%d']

        Step 2: Split initial dateformat by c-code groups

        Step 3: For token from previous step:
            If token is not a supported c-code, then treat the token as a literal and quote it.
                (e.g Year: -> "Year:")
            else:
                1) Get individual codes from token
                2) Replace each code with respective Postgres Code
                3) Recreate group by joining all results from previous step with an empty literal.
                    (e.g %Y%m%d -> YYYY""MM""DD

    .. note:: We use double quotes to force TO_CHAR interpret an empty string as literals, this way
            continuous date codes yield the correct value as intended
            For example having '%y%Y':
              TO_CHAR(cast('2022-01-01' as date), 'YY""YYYY') will generate '222022' (correct)
              vs.
              TO_CHAR(cast('2022-01-01' as date), 'YYYYYY') will generate '202222' (incorrect)

    """

    codes_base_pattern = '|'.join(_SUPPORTED_C_STANDARD_CODES)
    grouped_codes_matches = re.findall(pattern=rf"(?P<codes>(?:{codes_base_pattern})+)", string=date_format)
    if not grouped_codes_matches:
        # return it as a literal
        date_format = date_format.replace('"', r'\"')
        return f'"{date_format}"'

    tokenized_c_codes = sorted(set(grouped_codes_matches), key=len, reverse=True)
    single_c_code_regex = re.compile(rf'{codes_base_pattern}')

    new_date_format_tokens = []
    all_tokens = re.split(pattern=rf"({'|'.join(tokenized_c_codes)})", string=date_format)
    for token in all_tokens:
        # empty space
        if not token:
            continue

        # is a literal
        if token not in tokenized_c_codes:
            formatted_literal = token.replace('"', r'\"')
            new_date_format_tokens.append(f'"{formatted_literal}"')
            continue

        # get individual codes from the group
        codes_to_replace = single_c_code_regex.findall(token)
        replaced_codes = []
        for to_repl_code in codes_to_replace:
            # get correspondent postgres code if it exists, otherwise include to_repl_code unchanged.
            replaced_codes.append(_C_STANDARD_CODES_X_POSTGRES_DATE_CODES.get(to_repl_code, to_repl_code))

        new_date_format_tokens.append('""'.join(replaced_codes))

    return ''.join(new_date_format_tokens)


def _parse_c_code_to_athena_code(date_format: str) -> str:
    """
    Parses a date format string, and return a string that's compatible with Athena's date_format() function.

    Some python format codes are different on Athena, or might even raise an error. Such codes are converted
    or escaped; using the returned string with Athena's date_format will never raise an error.

    Does not support '%V', but besides that all codes in CODES_SUPPORTED_IN_ALL_DIALECTS are guaranteed to
    be correctly represented in the returned format string. If date_format contains codes that are not
    supported, then the returned format might evaluate to a different string than the original date_format
    would with python's strftime() function.
    """
    # Athena code spec: https://trino.io/docs/current/functions/datetime.html#mysql-date-functions
    # Python code spec: https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
    c_code_to_athena_mapping = {
        # if there is no code, i.e. a lone '%' that's not followed by an alphabetic character, then that
        # should yield '%'. So we want to generate '%%'
        '%': '%%',
        # List supported codes that are not in CODES_SUPPORTED_IN_ALL_DIALECTS
        '%f': '%f',
        # Actual mappings
        '%A': '%W',
        '%M': '%i',
        '%B': '%M',
        '%F': '%Y-%m-%d',
        '%R': '%H:%i',
        # List not-supported codes that are in CODES_SUPPORTED_IN_ALL_DIALECTS and will trigger an error if
        # not escaped. See also inline comments below.
        '%V': '%%V'
    }
    codes_to_consider = set(string.ascii_letters + '%')
    result = []
    i = 0
    while i < len(date_format):
        current = date_format[i]
        i += 1
        if current != '%':
            # This is not the start of a code sequence, just a regular character.
            result.append(current)
        else:
            # This is the start of a code sequence.
            if i < len(date_format) and date_format[i] in codes_to_consider:
                current = current + date_format[i]
                i += 1
            if current not in CODES_SUPPORTED_IN_ALL_DIALECTS and current not in c_code_to_athena_mapping:
                # If Athena encounters a code it does not support, then it will do either of:
                # 1) raise an error for a number of specifically not supported codes (%D, %U, %u, %V, %w, %X)
                # 2) only include the second character in the result, but strip the '%' off.
                # The general behaviour of strftime() in python is to just include the ignored code,
                # including the leading '%'. We want to mimic that by always escaping unknown codes.
                result.append(f'%{current}')
            else:
                # map c-codes to athena specific codes
                current = c_code_to_athena_mapping.get(current, current)
                result.append(current)
    return ''.join(result)


def _split_format_string_V(date_format: str) -> List[str]:
    """
    Split the date format string in a list of format strings. Each occurrence of the format code `%V` will
    be a separate format string in the resulting list, otherwise the format string is not split up.
    See inline comments of parse_c_code_to_athena_expression() on why one would like to do this.

    Examples:
    _split_format_string_V('%Y-%m-%d') == ['%Y-%m-%d']                          # zero '%V': no splits
    _split_format_string_V('xxx%V-%m-%%V%V') == ['xxx', '%V', '-%m-%%V', '%V']  # two '%V': split in four

    :param date_format: C date format string.
    :return: list of sub format strings.
    """
    result = []
    current_date_format = []
    i = 0
    while i < len(date_format):
        current = date_format[i]
        i += 1
        if current != '%' or i == len(date_format):
            current_date_format.append(current)
        else:
            # This is the start of a code sequence.
            code = current + date_format[i]
            i += 1
            if code != '%V':
                current_date_format.append(code)
            else:
                if current_date_format:
                    result.append(''.join(current_date_format))
                    current_date_format = []
                result.append('%V')
    if current_date_format or not result:
        result.append(''.join(current_date_format))
    return result


def parse_c_code_to_athena_expression(series_expression: Expression, date_format: str) -> Expression:
    """
    Parses a date format string, and return an Expression that represents the formatted date/timestamp/time.
    Depending on the date_format, the expression might combine calls to the Athena functions
    `date_format()` and `format_datetime()`.

    Codes in CODES_SUPPORTED_IN_ALL_DIALECTS are guaranteed to be correctly represented in the returned
    expression. If date_format contains codes that are not in that set, then the returned format might
    evaluate to a different string than the original date_format would with python's strftime() function.

    :param series_expression: The expression of the Date/Time/Timestamp/Timedelta Series that needs to be
        formatted
    :param date_format: C date format string
    """
    # Athena docs
    # format_datetime(): https://trino.io/docs/current/functions/datetime.html#java-date-functions
    # date_format():     https://trino.io/docs/current/functions/datetime.html#mysql-date-functions

    # Athena's date_format() supports most of the formatting codes that we are interested in.
    # Most non-supported codes can be directly translated to a supported one or a combination of supported
    # codes. Example: date_format() doesn't support '%R', but it does support '%H:%i', so we translate that.
    # These translations are handle in _parse_c_code_to_athena_code().
    # There is one exception to this, the code '%V' cannot be handled by date_format(), and there is no
    # alternative supported code either. However, a different function, format_datetime(), can return the
    # value of '%V'`. Therefore, we split the date_format string into parts: strings made up of '%V', and
    # all other strings. For the parts that contain just '%V' we'll call format_datetime, for all other parts
    # we'll call date_format. The results are all concatenated together.
    split_date_formats = _split_format_string_V(date_format)
    expressions: List[Expression] = []
    for sub_df in split_date_formats:
        if sub_df == '%V':
            expression = Expression.construct(
                "format_datetime({}, 'ww')", series_expression
            )
        else:
            parsed_format_str = _parse_c_code_to_athena_code(sub_df)
            if not parsed_format_str:
                expression = Expression.string_value('')
            else:
                expression = Expression.construct(
                    'date_format({}, {})', series_expression, Expression.string_value(parsed_format_str)
                )
        expressions.append(expression)
    return join_expressions(expressions, ' || ')


def parse_c_code_to_bigquery_code(date_format: str) -> str:
    """
    Replaces all '%S.%f' with '%E6S', since BigQuery does not support microseconds c-code.
    """
    # See BigQuery docs for more information:
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/format-elements#format_elements_date_time
    if '%S.%f' in date_format:
        date_format = re.sub(r'%S\.%f', '%E6S', date_format)
    return date_format


def warn_non_supported_format_codes(date_format: str):
    """
    Checks that all formatting codes in date_format are listed in CODES_SUPPORTED_IN_ALL_DIALECTS or are
    part of a string that's listed in STRINGS_SUPPORTED_IN_ALL_DIALECTS.

    If one or more non-listed codes are found, a UserWarning is emitted.
    """
    unsupported_c_codes = set()
    i = 0
    while i < len(date_format):
        current = date_format[i]
        i += 1
        if current != '%':  # This is not the start of a code sequence, just a regular character.
            continue

        # This is the start of a code sequence.
        # See if any of the STRINGS_SUPPORTED_IN_ALL_DIALECTS start at this character
        for supported_string in STRINGS_SUPPORTED_IN_ALL_DIALECTS:
            # Get a string from date_format with the same length as supported_string starting at the current
            # position.
            sub_str_start = i - 1
            sub_str_end = sub_str_start + len(supported_string)
            sub_str = date_format[sub_str_start:sub_str_end]
            if sub_str == supported_string:
                # Match: we know this string is good, and can skip to the end of it.
                i = sub_str_end
                break  # skip the 'else:' clause of this for loop
        else:
            # If we get here, that means that none of the strings in STRINGS_SUPPORTED_IN_ALL_DIALECTS
            # start at the current character.
            # See if the code that starts here is listed in CODES_SUPPORTED_IN_ALL_DIALECTS
            if i < len(date_format):
                current = current + date_format[i]
                i += 1
            if current not in CODES_SUPPORTED_IN_ALL_DIALECTS:
                unsupported_c_codes.add(current)
    if unsupported_c_codes:
        message = f'These formatting codes are not generally supported: ' \
                  f'{", ".join(sorted(unsupported_c_codes))}.' \
                  f'They might not work reliably on some or all database platforms..'
        warnings.warn(
            message=message,
            category=UserWarning,
        )



"""
Copyright 2021 Objectiv B.V.
"""
__version__ = '0.0.27.dev4'

from bach.dataframe import DataFrame, DataFrameOrSeries, ColumnNames, SortColumn, \
    get_series_type_from_dtype
from bach.series import *
from bach.display_formats import display_sql_as_markdown

# TODO: check. Do we need to generate docs for this at this point?
from_table = DataFrame.from_table
from_model = DataFrame.from_model
from_pandas = DataFrame.from_pandas



"""
Copyright 2021 Objectiv B.V.
"""
from bach.series.series import Series, value_to_series
from bach.series.series_numeric import SeriesAbstractNumeric, SeriesInt64, SeriesFloat64
from bach.series.series_boolean import SeriesBoolean
from bach.series.series_uuid import SeriesUuid
from bach.series.series_json import SeriesJson, SeriesJsonPostgres
from bach.series.series_string import SeriesString
from bach.series.series_datetime import \
    SeriesAbstractDateTime, SeriesDate, SeriesTime, SeriesTimestamp, SeriesTimedelta
from bach.series.series_multi_level import SeriesAbstractMultiLevel, SeriesNumericInterval
from bach.series.series_list import SeriesList
from bach.series.series_dict import SeriesDict



"""
Copyright 2022 Objectiv B.V.
"""
from typing import Any, Tuple, TYPE_CHECKING, Dict, Optional, Mapping, Union

from sqlalchemy.engine import Dialect

from bach.series import Series
from bach.expression import Expression, join_expressions
from bach.types import DtypeOrAlias, get_series_type_from_dtype, StructuredDtype, Dtype, validate_dtype_value
from sql_models.constants import DBDialect
from sql_models.model import Materialization
from sql_models.util import DatabaseNotSupportedException, is_bigquery


if TYPE_CHECKING:
    from bach import DataFrameOrSeries


class SeriesDict(Series):
    """
    A Series that represents a dictionary-like type and its specific operations.
    On BigQuery this is backed by the STRUCT data type. On other databases this type is not supported.


    **instance_dtype:**
    An instance of this class expects an ``instance_dtype`` that's a dictionary. With each key being the
    name of a field, and the value being an instance-dtype describing that field. The values represented by
    such a Series must then be dictionaries with the exact same field names as the instance-dtype, and the
    specified data types for the fields.

    Example -
    BigQuery db_dtype: ``'STRUCT<a INT64, b ARRAY[FLOAT64], c <STRUCT x BOOL, y: INT64>>'``
    instance_dtype: ``{'a': 'int64', b: ['float64'], 'c': {'x': 'bool', 'y': 'int64'}``


    **Database support and types**

    * Postgres: not supported. Use :class:`SeriesJson` for similar functionality.
    * BigQuery: utilizes the 'STRUCT' database type.
    """
    dtype = 'dict'
    dtype_aliases: Tuple[DtypeOrAlias, ...] = tuple()
    supported_db_dtype = {
        # We support BigQuery STRUCT type, but the exact db_dtype depends on what kind of data the struct
        # holds (e.g. 'STRUCT<x INT64>'). So we don't define a static db_dtype that we support
        DBDialect.BIGQUERY: None
        # Postgres is not supported, thus not listed here
    }
    supported_value_types = (dict, )

    # This series does not support casting from other dtypes
    supported_source_dtypes: Tuple[str, ...] = tuple()

    @classmethod
    def supported_value_to_literal(
            cls,
            dialect: Dialect,
            value: Dict[str, Any],
            dtype: StructuredDtype
    ) -> Expression:
        # validate early, and help mypy
        cls.assert_engine_dialect_supported(dialect)
        if not isinstance(dtype, dict):
            raise ValueError(f'Dtype should be type dict. Type(dtype): {type(dtype)}')
        validate_dtype_value(static_dtype=cls.dtype, instance_dtype=dtype, value=value)

        sub_exprs = []
        for key, item in value.items():
            sub_dtype = dtype[key]
            series_type = get_series_type_from_dtype(sub_dtype)
            sub_val = series_type.value_to_expression(
                dialect=dialect,
                value=item,
                dtype=sub_dtype
            )
            sub_expr = Expression.construct('{} as {}', sub_val, Expression.identifier(key))
            sub_exprs.append(sub_expr)
        return Expression.construct('STRUCT({})', join_expressions(expressions=sub_exprs, join_str=', '))

    @classmethod
    def from_value(
        cls,
        base: 'DataFrameOrSeries',
        value: Dict[str, Any],
        name: str = 'new_series',
        dtype: Optional[StructuredDtype] = None
    ) -> 'SeriesDict':
        """
        Create an instance of this class, that represents a column with the given dict as value.
        The given base Series/DataFrame will be used to set the engine, base_node, and index.

        :param base:    The DataFrame or Series that the internal parameters are taken from
        :param value:   The value that this constant Series will have. Cannot be null.
        :param name:    The name that it will be known by (only for representation)
        :param dtype:   instance dtype, mandatory. Should be a dict, describing the structure of value.
        """
        # We override the parent class here to allow using Series as sub-values in a dict

        # validate early, and help mypy
        cls.assert_engine_dialect_supported(base.engine)
        if not isinstance(dtype, dict):
            raise ValueError(f'Dtype should be type dict. Type(dtype): {type(dtype)}')
        if value is None:
            raise ValueError(f'None values are not supported in from_value() by this class.')
        validate_dtype_value(static_dtype=cls.dtype, instance_dtype=dtype, value=value)

        sub_exprs = []
        for key, item in value.items():
            if isinstance(item, Series):
                if item.base_node != base.base_node:
                    raise ValueError(f'When constructing a struct from existing Series. '
                                     f'The Series.base_node must match the base.base_node. '
                                     f'Key with incorrect base-node: {key}')
                series = item
            else:
                sub_dtype = dtype[key]
                series_type = get_series_type_from_dtype(sub_dtype)
                series = series_type.from_value(
                    base=base,
                    value=item,
                    name=key,
                    dtype=sub_dtype
                )
            sub_expr = Expression.construct('{} as {}', series.expression, Expression.identifier(key))
            sub_exprs.append(sub_expr)
        expr = Expression.construct('STRUCT({})', join_expressions(expressions=sub_exprs, join_str=', '))
        result = cls.get_class_instance(
            engine=base.engine,
            base_node=base.base_node,
            index=base.index,
            name=name,
            expression=expr,
            group_by=None,
            order_by=[],
            instance_dtype=dtype
        )
        return result

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        # Even when casting from a dict to a dict, things will be troublesome if instance_dtype doesn't
        # match. So just raise an error.
        raise Exception('Cannot cast a value to a dictionary.')

    @property
    def elements(self) -> 'DictAccessor':
        """ Accessor to interact with the elements in the dictionary. """
        return DictAccessor(self)

    def materialize(
            self,
            node_name='manual_materialize',
            limit: Any = None,
            distinct: bool = False,
            materialization: Union[Materialization, str] = Materialization.CTE
    ):
        raise Exception(f'{self.__class__.__name__} cannot be materialized.')


class DictAccessor:
    def __init__(self, series: SeriesDict):
        self._series = series
        # Below checks on self._series.instance_dtype are not strictly needed, but it helps mypy. We checked
        # this in Series.__init__(), by calling validate_instance_dtype()
        if not isinstance(self._series.instance_dtype, dict):
            raise Exception('Found type: {self._series.instance_dtype}')
        self._instance_dtype: Dict[str, Any] = self._series.instance_dtype

    def __getitem__(self, key: str) -> 'Series':
        """
        Get an item from the dictionary.
        :param key: name of the field to get.
        """
        instance_dtype = self._instance_dtype
        if key not in instance_dtype:
            raise ValueError(f'Invalid key: {key}. '
                             f'Available keys: {sorted(instance_dtype.keys())}')
        expression = Expression.construct('{}.{}', self._series, Expression.identifier(key))
        sub_dtype = instance_dtype[key]
        if isinstance(sub_dtype, Dtype):
            new_dtype = sub_dtype
            return self._series \
                .copy_override_dtype(dtype=new_dtype) \
                .copy_override(expression=expression)
        elif isinstance(sub_dtype, dict):
            return self._series \
                .copy_override_type(SeriesDict, instance_dtype=sub_dtype) \
                .copy_override(expression=expression)
        elif isinstance(sub_dtype, list):
            from bach import SeriesList
            return self._series \
                .copy_override_type(SeriesList, instance_dtype=sub_dtype) \
                .copy_override(expression=expression)
        else:
            raise Exception(f'Unsupported structural type: {sub_dtype}')



"""
Copyright 2022 Objectiv B.V.
"""
from enum import Enum
from typing import cast, List, Union
from bach import (
    SeriesAbstractNumeric, Series, DataFrame,
    SeriesBoolean, SeriesString, SeriesNumericInterval,
)
from bach.expression import Expression
import numpy

_RANGE_ADJUSTMENT = 0.001  # Pandas.cut currently uses 1%


class CutMethod(Enum):
    PANDAS = 'pandas'
    BACH = 'bach'


class CutOperation:
    """
    In order to instantiate this class you should provide the following params:
    series: a numerical series to be segmented into bins
    bins: number of equal-width bins the series will be divided into.
    right: indicates if bin ranges should include the rightmost edge (lower bound).
    ignore_index: if true, original index will be dropped
    include_empty_bins: if true, it will return bins that contain no values in series.
    method: Method to use for calculating ranges. Supported values:
       - "pandas": Yields same intervals as ``pandas.cut`` by applying lower/upper boundary adjustments
            based on minimum and maximum values.
       - "bach": Intervals are not adjusted, initial/last interval's boundaries are changed to include both
        lower and upper boundaries. If right == True, the first interval from all bins is changed to '[]',
        else the last interval from all bins is changed to '[]'.

    returns a new Numerical Series

    Example:
        CutOperation(series=s1, bins=3)()
    """
    series: SeriesAbstractNumeric
    bins: int
    right: bool
    include_empty_bins: bool
    ignore_index: bool
    method: CutMethod

    RANGE_SERIES_NAME = 'range'

    def __init__(
        self,
        series: SeriesAbstractNumeric,
        bins: int,
        right: bool = True,
        ignore_index: bool = True,
        include_empty_bins: bool = False,
        method: str = 'pandas',
    ) -> None:
        self.series = series
        self.bins = bins
        self.right = right
        self.ignore_index = ignore_index
        self.include_empty_bins = include_empty_bins
        self.method = CutMethod(method)

    def __call__(self) -> SeriesNumericInterval:
        """
        Merges self.series with its correspondent bucket range.

        returns a series containing each bin range/interval per value,
        if self.include_empty_bins is True, ranges without values will be considered.
        (initial series will be contained as index)
        """
        bucket_properties_df = self._calculate_bucket_properties()
        range_df = self._calculate_bucket_ranges(bucket_properties_df)

        final_index_keys = []
        if not self.ignore_index:
            final_index_keys = list(self.series.index.keys())
        final_index_keys += [self.series.name]

        df = self.series.to_frame().reset_index(drop=self.ignore_index)

        left_df = df if not self.include_empty_bins else range_df
        right_df = range_df if not self.include_empty_bins else df
        how = 'inner' if not self.include_empty_bins else 'left'

        fake_merge = left_df.merge(right_df, how='cross')
        if self.right:
            mask = (
                (fake_merge[self.series.name] > fake_merge['lower_bound'])
                & (fake_merge[self.series.name] <= fake_merge['upper_bound'])
            )
        else:
            mask = (
                 (fake_merge[self.series.name] >= fake_merge['lower_bound'])
                 & (fake_merge[self.series.name] < fake_merge['upper_bound'])
            )

        if self.method == CutMethod.BACH:
            mask |= (
                (fake_merge[self.series.name] == fake_merge['lower_bound' if self.right else 'upper_bound'])
                & (fake_merge['bucket'] == (1 if self.right else self.bins))
            )

        df = left_df.merge(right_df, how=how, on=mask)

        df = df.set_index(keys=final_index_keys)

        range_series = SeriesNumericInterval.from_value(
            base=df,
            value={
                'lower': df['lower_bound'],
                'upper': df['upper_bound'],
                'bounds': df['bounds'],
            },
            name=self.RANGE_SERIES_NAME,
        )
        return cast(SeriesNumericInterval, range_series)

    @property
    def bounds(self) -> str:
        return "'(]'" if self.right else "'[)'"

    def _calculate_bucket_properties(self) -> 'DataFrame':
        """
        Calculates the following properties that are required
        for the bucket calculations and range adjustments:
        * min: Lower boundary for width_bucket. Only for method == 'pandas': series.min - min_adjustment

        * max: Upper boundary for width_bucket. Only for method == 'pandas': series.max - max_adjustment.

        * step: the size of each bucket range

        if method == 'pandas':
            * min_adjustment: the value to be subtracted to series.min
                if series.min == series.max,
                minimum value is decreased by RANGE_ADJUSTMENT * abs(series.min)
                if series.min > 0, otherwise just by the RANGE_ADJUSTMENT

            * max_adjustment: the value to be added to series.max
                if series.min == series.max,
                maximum value is increased by RANGE_ADJUSTMENT * abs(series.max)
                if series.max > 0, otherwise just by the RANGE_ADJUSTMENT

            * bin_adjustment: value used to adjust lower/upper bounds of the final bucket ranges.
               Based on (max - min) * RANGE_ADJUSTMENT


        returns a DataFrame containing the following calculated columns:
        * {self.series.name}_min: the final minimum value of the series with applied adjustments
        * {self.series.name}_max: the final maximum value of the series with applied adjustments
        * step: The equal-length width of each bucket.
        if method == 'pandas':
            * bin_adjustment: the adjustment to be applied to the lower bound of the first range or
            the upper bound of the last range. This is based on self.right

        ** Adjustments are needed in order to have similar bin intervals as in Pandas
        """
        df = self.series.to_frame().reset_index(drop=True)

        properties_df = df.agg(['min', 'max'])
        min_name = f'{self.series.name}_min'
        max_name = f'{self.series.name}_max'

        if self.method == CutMethod.BACH:
            properties_df['step'] = (properties_df[max_name] - properties_df[min_name]) / self.bins
            return properties_df[[min_name, max_name, 'step']]

        properties_df['min_adjustment'] = self._calculate_pandas_adjustments(
            to_adjust=properties_df[min_name], compare_with=properties_df[max_name],
        )
        properties_df['max_adjustment'] = self._calculate_pandas_adjustments(
            to_adjust=properties_df[max_name], compare_with=properties_df[min_name],
        )

        # need to adjust both min and max with the prior calculated adjustment
        # this is mainly to avoid the case both min and max are equal
        properties_df[min_name] = properties_df[min_name] - properties_df['min_adjustment']
        properties_df[max_name] = properties_df[max_name] + properties_df['max_adjustment']

        diff_min_max = properties_df[max_name] - properties_df[min_name]
        # value used for expanding start/end bound
        properties_df['bin_adjustment'] = diff_min_max * _RANGE_ADJUSTMENT
        properties_df['step'] = diff_min_max / self.bins

        final_properties = [min_name, max_name, 'bin_adjustment', 'step']
        return properties_df[final_properties]

    def _calculate_pandas_adjustments(
        self, to_adjust: 'Series', compare_with: 'Series'
    ) -> 'Series':
        """
        calculates adjustment when to_adjust == compare_with.
        If both are equal, calculations based on RANGE_ADJUSTMENT will be performed:
        * if to_adjust != 0: RANGE_ADJUSTMENT * abs(to_adjust) else RANGE_ADJUSTMENT

        returns a Series with the calculated adjustment
        """
        case_stmt = (
            f'case when {{}} = {{}} then\n'
            f'case when {{}} != 0 then {_RANGE_ADJUSTMENT} * abs({{}}) else {_RANGE_ADJUSTMENT} end\n'
            f'else 0 end'
        )
        return to_adjust.copy_override(
            expression=Expression.construct(
                case_stmt,
                compare_with,
                *[to_adjust] * 3,
            )
        )

    def _calculate_bucket_ranges(self, bucket_properties_df: 'DataFrame') -> 'DataFrame':
        """
        Calculates upper and lower bound for each bucket.
         * bucket (integer 1 to N)
         * lower_bound (float). if method == 'pandas', adjustments might be performed based on self.right
         * upper_bound (float). if method == 'pandas', adjustments might be performed based on self.right

        return a DataFrame with the calculated series from above
        """

        # self.series might not have data for all buckets, we need to actually generate the series
        import pandas
        from bach.dataframe import DataFrame
        range_df = DataFrame.from_pandas(
            engine=self.series.engine,
            df=pandas.DataFrame(data={'bucket': range(1, self.bins + 1)})
        ).reset_index(drop=True)

        range_df = range_df.merge(bucket_properties_df, how='cross')

        # lower_bound = (bucket - 1) *  step + min
        range_df['lower_bound'] = (
            (range_df.bucket - 1) * range_df['step'] + range_df[f'{self.series.name}_min']
        )
        # upper_bound = bucket * step + min
        range_df['upper_bound'] = range_df.bucket * range_df['step'] + range_df[f'{self.series.name}_min']

        bounds_stmt = self.bounds
        if self.method == CutMethod.BACH:
            bounds_stmt = (
                f"case when bucket = {1 if self.right else self.bins} then '[]' else {self.bounds} end"
            )
        elif self.method == CutMethod.PANDAS:
            if self.right:
                case_stmt = f'case when bucket = 1 then {{}} - {{}} else {{}} end'
                bound_to_adjust = 'lower_bound'
            else:
                case_stmt = f'case when bucket = {self.bins} then {{}} + {{}} else {{}} end'
                bound_to_adjust = 'upper_bound'

            # expand the correspondent boundary
            range_df[bound_to_adjust] = range_df[bound_to_adjust].copy_override(
                expression=Expression.construct(
                    case_stmt,
                    range_df[bound_to_adjust],
                    Series.as_independent_subquery(bucket_properties_df['bin_adjustment']),
                    range_df[bound_to_adjust],
                ),
            )

        range_df = range_df.materialize(node_name='bin_ranges')
        range_df['bounds'] = range_df['lower_bound'].copy_override(
            name='bounds',
            expression=Expression.construct(bounds_stmt),
        ).copy_override_type(SeriesString)

        return range_df[['bucket', 'lower_bound', 'upper_bound', 'bounds']]


class QCutOperation:
    """
    In order to instantiate this class you should provide the following params:
    series: A numerical series
    q: The number of quantiles or list of quantiles to be calculated

    returns a new Series containing the quantile ranges per each value on the series.

    Example:
        QCutOperation(s1, q=4)()   # will calculated quantiles `0, 0.25, 0.5, 0.75, 1`
        or
        QCutOperation(s1, q=[0.25, 0.5, 0.75])()
    """

    series: SeriesAbstractNumeric
    quantiles: List[float]

    RANGE_SERIES_NAME = 'q_range'

    def __init__(self, series: SeriesAbstractNumeric, q: Union[int, List[float]]) -> None:
        self.series = series
        self.quantiles = q if isinstance(q, list) else numpy.linspace(0, 1, q + 1).tolist()

    def __call__(self, *args, **kwargs) -> 'SeriesNumericInterval':
        """
        Gets the quantile range per bucket and assigns the correct range to each value. If the value
        is not contained in any range, then it will be null.
        """
        df = self.series.to_frame().reset_index(drop=True)

        if len(self.quantiles) == 1:
            # need at least 2 quantiles for a range
            df[self.RANGE_SERIES_NAME] = SeriesNumericInterval.from_value(
                base=df, value=None, name=self.RANGE_SERIES_NAME,
            )
            df = df.set_index(self.series.name)
            return cast(SeriesNumericInterval, df[self.RANGE_SERIES_NAME])

        quantile_ranges = self._get_quantile_ranges()
        fake_merge = df.merge(quantile_ranges, how='cross')
        mask = (
            (fake_merge[self.series.name] > fake_merge['lower_bound'])
            & (fake_merge[self.series.name] <= fake_merge['upper_bound'])
        )
        mask = mask.copy_override_type(SeriesBoolean)
        df = df.merge(quantile_ranges, how='left', on=mask)
        df[self.RANGE_SERIES_NAME] = SeriesNumericInterval.from_value(
            base=df,
            value={
                'lower': df['lower_bound'],
                'upper': df['upper_bound'],
                'bounds': df['bounds'],
            },
            name=self.RANGE_SERIES_NAME,
        )

        df = df.set_index(self.series.name)
        return cast(SeriesNumericInterval, df[self.RANGE_SERIES_NAME])

    def _get_quantile_ranges(self) -> 'DataFrame':
        """
        Calculates the corresponding ranges per each quantile bucket.

        The following series are calculated:
        * lower_bound: for calculating the lower bound of each range it is required to calculate
            all requested quantiles from the series. The lowest bound is adjusted since it might be
            included in the dataset.
        * upper_bound: is the lower bound from the next range that follows the current one.
            If current lower bound is the result of the largest quantile, the upper bound will be null.

        Returns a dataframe

        .. note::
            The adjustment performed on the lowest bound is done only to resemble Panda's implementation.
            Current implementation might generate wrong ranges in very extreme edge cases,
            such as when _RANGE_ADJUSTMENT is considerably larger than the first quantile result.
            Therefore, be aware of this scenario.
            Current implementation might go into a discussion in the future.
        """
        q_result = self.series.quantile(q=self.quantiles).copy_override(name='q_result')

        quantile_ranges_df = q_result.to_frame()

        # lowest calculated quantile might be also be in the dataset, therefore
        # we need to extend the lowest bound
        # Be aware that this adjustment might generate errors when
        # 0 < lowest_quantile < RANGE_ADJUSTMENT
        min_quantile = min(self.quantiles)
        mask = quantile_ranges_df.index['quantile'] == min_quantile
        quantile_ranges_df.loc[mask, 'q_result'] = quantile_ranges_df['q_result'] - _RANGE_ADJUSTMENT

        quantile_ranges_df['bounds'] = '(]'

        # some quantiles might have the same result, we need to avoid having overlapped ranges
        quantile_ranges_df = quantile_ranges_df.drop_duplicates(ignore_index=True)

        quantile_ranges_df['lower_bound'] = quantile_ranges_df['q_result'].copy()
        window = quantile_ranges_df.sort_values(by='q_result').groupby().window()
        quantile_ranges_df['upper_bound'] = quantile_ranges_df['q_result'].window_lead(window=window)

        quantile_ranges_df = quantile_ranges_df[['lower_bound', 'upper_bound', 'bounds']]
        return quantile_ranges_df.materialize(node_name='q_range_calculation')



"""
Copyright 2022 Objectiv B.V.
"""
from typing import Any, Tuple, List, Union, TYPE_CHECKING, Optional, Mapping, Sequence

from sqlalchemy.engine import Dialect

from bach.series import Series
from bach.expression import Expression, join_expressions
from bach.types import DtypeOrAlias, get_series_type_from_dtype, StructuredDtype, Dtype, validate_dtype_value
from sql_models.constants import DBDialect
from sql_models.util import DatabaseNotSupportedException, is_bigquery


if TYPE_CHECKING:
    from bach import SeriesInt64, DataFrameOrSeries


class SeriesList(Series):
    """
    A Series that represents a list/array-like type and its specific operations.
    On BigQuery this is backed by the ARRAY data type. On other databases this type is not yet supported.


    **instance_dtype:**
    An instance of this class expects an ``instance_dtype`` that's a list with a single item in it, the item
    itself being an instance dtype. All values represented by such a Series must have the dtype of the item
    in the list.

    Example - instance_dtype for a list consisting of floats.
    BigQuery db_dtype: ``'ARRAY[FLOAT64]'``
    instance_dtype: ``['float64']``

    **Database support and types**

    * Postgres: not supported. Use :class:`SeriesJson` for similar functionality.
    * BigQuery: utilizes the 'ARRAY' database type.
    """
    dtype = 'list'
    dtype_aliases: Tuple[DtypeOrAlias, ...] = tuple()
    supported_db_dtype = {
        # We support BigQuery ARRAY type, but the exact db_dtype depends on what kind of data the array holds
        # (e.g. 'ARRAY<INT64>'). So we don't define a static db_dtype that we support
        DBDialect.BIGQUERY: None
        # Postgres is not supported, thus not listed here
    }

    supported_value_types = (list, )
    # This series does not support casting from other dtypes
    supported_source_dtypes: Tuple[str, ...] = tuple()

    @classmethod
    def supported_value_to_literal(
        cls,
        dialect: Dialect,
        value: List[Any],
        dtype: StructuredDtype
    ) -> Expression:
        cls.assert_engine_dialect_supported(dialect)

        # validate early, and help mypy
        if not isinstance(dtype, list):
            raise ValueError(f'Dtype should be type list. Type(dtype): {type(dtype)}')
        validate_dtype_value(static_dtype=cls.dtype, instance_dtype=dtype, value=value)

        sub_dtype = dtype[0]
        series_type = get_series_type_from_dtype(sub_dtype)
        sub_exprs = [
            series_type.value_to_expression(dialect=dialect, value=item, dtype=sub_dtype)
            for item in value
        ]

        return cls._sub_expressions_to_expression(
            dialect=dialect,
            sub_dtype=sub_dtype,
            sub_expressions=sub_exprs
        )

    @classmethod
    def from_value(
        cls,
        base: 'DataFrameOrSeries',
        value: List[Any],
        name: str = 'new_series',
        dtype: Optional[StructuredDtype] = None
    ) -> 'SeriesList':
        """
        Create an instance of this class, that represents a column with the given list as value.
        The given base Series/DataFrame will be used to set the engine, base_node, and index.

        :param base:    The DataFrame or Series that the internal parameters are taken from
        :param value:   The value that this constant Series will have. Cannot be null.
        :param name:    The name that it will be known by (only for representation)
        :param dtype:   instance dtype, mandatory. Should be a list with one item in it, the item being the
                        instance dtype of what is contained in the list
        """
        # We override the parent class here to allow using Series as sub-values in a list

        cls.assert_engine_dialect_supported(base.engine)
        # validate early, and help mypy
        if not isinstance(dtype, list):
            raise ValueError(f'Dtype should be type list. Type(dtype): {type(dtype)}')
        if value is None:
            raise ValueError(f'None values are not supported in from_value() by this class.')
        validate_dtype_value(static_dtype=cls.dtype, instance_dtype=dtype, value=value)

        sub_dtype = dtype[0]
        series_type = get_series_type_from_dtype(sub_dtype)
        sub_exprs = []
        for i, item in enumerate(value):
            if isinstance(item, Series):
                if item.base_node != base.base_node:
                    raise ValueError(f'When constructing a SeriesList from existing Series. '
                                     f'The Series.base_node must match the base.base_node. '
                                     f'Index with incorrect base-node: {i}')
                series = item
            else:
                series = series_type.from_value(
                    base=base,
                    value=item,
                    name=f'item_{i}',
                    dtype=sub_dtype
                )
            sub_exprs.append(series.expression)

        expression = cls._sub_expressions_to_expression(
            dialect=base.engine.dialect,
            sub_dtype=sub_dtype,
            sub_expressions=sub_exprs
        )
        result = cls.get_class_instance(
            engine=base.engine,
            base_node=base.base_node,
            index=base.index,
            name=name,
            expression=expression,
            group_by=None,
            order_by=[],
            instance_dtype=dtype
        )
        return result

    @staticmethod
    def _sub_expressions_to_expression(
        dialect: Dialect,
        sub_dtype: StructuredDtype,
        sub_expressions: Sequence[Expression]
    ) -> Expression:
        """ Internal function: create an array expression from a list of expressions """
        series_type = get_series_type_from_dtype(sub_dtype)
        if not sub_expressions:
            # Special case: empty array, we need to tell the database the type of the array
            sub_db_dtype = series_type.get_db_dtype(dialect)
            if sub_db_dtype is None:
                # We expect this if the sub-types are structured types themselves
                raise ValueError("Empty lists of structured types are not supported.")
            return Expression.construct(f'ARRAY<{sub_db_dtype}>[]', )

        return Expression.construct(
            '[{}]',
            join_expressions(expressions=sub_expressions, join_str=', ')
        )

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        # Even when casting from a list to a list, things will be troublesome if instance_dtype doesn't
        # match. So just raise an error.
        raise Exception('Cannot cast a value to a list.')

    @property
    def elements(self) -> 'ListAccessor':
        return ListAccessor(self)


class ListAccessor:
    def __init__(self, series: SeriesList):
        self._series = series

    def __getitem__(self, key: Union[int, slice]) -> 'Series':
        """
        Get an item from the list.
        :param key: integer indicating position in the list, 0-based. Slice is not yet supported
        """
        engine = self._series.engine
        if isinstance(key, int):
            if is_bigquery(engine):
                expr_str = f'{{}}[SAFE_OFFSET({key})]'
            else:
                raise DatabaseNotSupportedException(engine)

            expression = Expression.construct(expr_str, self._series)
            # help mypy. We checked this in Series.__init__(), by calling validate_instance_dtype()
            assert isinstance(self._series.instance_dtype, list)
            sub_dtype = self._series.instance_dtype[0]

            if isinstance(sub_dtype, Dtype):
                new_dtype = sub_dtype
                return self._series \
                    .copy_override_dtype(dtype=new_dtype) \
                    .copy_override(expression=expression)
            elif isinstance(sub_dtype, list):
                return self._series \
                    .copy_override(instance_dtype=sub_dtype, expression=expression)
            elif isinstance(sub_dtype, dict):
                from bach import SeriesDict
                return self._series \
                    .copy_override_type(SeriesDict, instance_dtype=sub_dtype) \
                    .copy_override(expression=expression)
            else:
                raise Exception(f'Unexpected type of sub_dtype. sub_dtype: {sub_dtype}')

        elif isinstance(key, slice):
            # TODO: should be easy on Postgres. Could be a lot of work on BigQuery for a nice solution.
            #   or we can just concat a lot of __getitem__s
            return self._series
        else:
            raise ValueError(f'Invalid key type: {type(key)}')

    def len(self) -> 'SeriesInt64':
        engine = self._series.engine
        if is_bigquery(engine):
            expr_str = 'ARRAY_LENGTH({})'
        else:
            raise DatabaseNotSupportedException(engine)
        from bach import SeriesInt64
        return self._series \
            .copy_override_type(SeriesInt64) \
            .copy_override(expression=Expression.construct(expr_str, self._series))
